/* Ilu.java */
/* Chris Jacobi, September 11, 1997 11:26 am PDT */

/*
 * Copyright (c) 1996, 1997 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: Ilu.java,v 1.19 1997/09/11 18:27:29 jacobi Exp $ */

 
package xerox.ilu;
import xerox.ilu.IluClassRep;
import xerox.ilu.IluInit;
import xerox.ilu.IluObject;
import xerox.ilu.IluSystemException;
import xerox.ilu.IluRT0;
import xerox.ilu.IluServer;

/**
 * Main application interface for ILU.<p>
 */
public final class Ilu {
        
    /**
     * Don't create instances 
     */
    private Ilu(){};
    
    /** 
     * Importing an object 
     */
    public static IluObject 
    objectFromSBH (
            java.lang.String sbh, 
            IluClassRep iluClass
            ) throws IluSystemException {
        return IluRT0.objectFromSBH(sbh, iluClass);
    }

    /**
     * Create true objects.
     * OBSOLETE.
     */
    public static void 
    registerTrueObject(
            java.lang.String ih,
            IluObject tobj, 
            IluServer server, 
            IluClassRep iluClass
            ) throws IluSystemException {
        registerTrueObject(ih, 
            tobj, server, iluClass, IluLifetimeArgs.iluLifetimeRemember
            );
    }

    /** 
     * Literals for the lifetime argument of registerTrueObject.
     */
    
    /** Requires ilu set a value depending on other arguments */
    public static final int unspec = IluLifetimeArgs.iluLifetimeUnspec;
    /** Remember obj to prevent java garbage collection. */
    public static final int remember = IluLifetimeArgs.iluLifetimeRemember;
    /** Forget obj to subject it to java garbage collection. */
    public static final int forget = IluLifetimeArgs.iluLifetimeForget;

    /**
     * Create true objects. <p>
     * Life time is important and must be determined.  Normally the ilu
     * object stays alive until the intermediate object is garbage
     * collected.  Use any of a large number of schemes to prevent
     * garbage collection. <p>
     * 
     * Return value is intermediate object.  Caller may prevent this
     * from being garbage collected. <p>
     *
     * @see IluLifetimeArgs
     * @see IluIHProposer
     *
     */
    public static java.lang.Object 
    registerTrueObject(
            java.lang.String ih,
            IluObject tobj, 
            IluServer server, 
            IluClassRep iluClass,
            int lifetime //0 gives default behaviour
            ) throws IluSystemException {
        return IluRT0.registerTrueObject(ih, tobj, server, iluClass, lifetime);
    }
    
    
    /** 
     * Returns the string binding handle.<br>
     * May return NIL if the object's server isn't exported through any
     * port; may return an invalid SBH if the cached one references a
     * closed port.
     */
    public static java.lang.String 
    sbhOfObject(IluObject obj) {
        return IluRT0.sbhOfObject(obj);
    }
    
    
    /** 
     * Returns OMG IIOP-specified IOR string for object. <br>
     * May return null if object is not exported through IIOP.
     */
    public static java.lang.String 
    iorOfObject(IluObject obj) {
        return IluRT0.iorOfObject(obj);
    }
  
  
    /** 
     * Destroy ILU-ness of true object.<p>
     *
     * Only for impatient applications,
     * Normal applications leave this to the garbage collector. <br>
     * Do NOT use from applets; error behavior is not yet defined. <br>
     * Do not use on surrogates.
     */
    public static void 
    destroyObject(IluObject obj) {
        IluRT0.destroyObject(obj);
    }
    
    
    /**
     * Generates a string that's unique over space and time.
     * A server with nothing better to use might call this to 
     * get an ID.
     */
    public static java.lang.String 
    inventID() {
        return IluRT0.inventID();
    }
    
    
    /**
     * Test whether object is alive.
     */
    public static void 
    ping(IluObject obj) {
        IluRT0.ping(obj);
    }
    

    /**
     * Makes sure Ilu is loaded.<p>
     * Idempotent.
     */
    public static void init(){};
    
    static {
       IluInit.init();
    }
  
} //Ilu
