/* IluOInt.java */
/* Chris Jacobi, September 18, 1997 5:36 pm PDT */

/*
 * Copyright (c) 1996, 1997 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluOInt.java,v 1.38 1997/09/19 01:42:17 jacobi Exp $ */

/*
 * This class helps in the mapping from real Java objects to kernel object 
 * See also IluJava_IluOInt.c 
 */

package xerox.ilu;
import java.util.Hashtable;
import xerox.basics.Consumer0;

/**
 * This class helps in the mapping from real Java objects to kernel object;
 * It resides in the middle.  It implements what the kernel considers a 
 * LSO. While applications might hold an IluOInt; none of the operations
 * here should be used by applications.
 */
public final class IluOInt extends IluWPBase {
    /* friends */ long yIluKernelObject = 0;
    /* friends */ long yServer = 0;//true AND surrogate
    /* friends */ long yOwnerKey = 0;
    /* friends */ xerox.ilu.IluServer jjServer = null; //currently only if true
    /* friends */ boolean veryInterest = false; //by kernel
    /* friends */ boolean retained = false; //by application
    /* friends */ xerox.ilu.IluReEnableFinalization reFinalizer = null;
    /* friends */ boolean destroyed = false;
    /* friends */ boolean ghost = false; //Used unprotected. Delays garbage
        //collection a turn.  Used conceptionally to un-enqueue finalizations
        //when a disguised ref is undisguised but finalization is already
        //enqueued from before undisguisation.  Since used unprotected, must
        //not participate in proof of memory safety.
    /* friends */ xerox.ilu.IluClassRep jjClassRep = null;
    /* friends */ xerox.ilu.IluObject jjJavaObject = null; 
        //in synch with peektab
    private static Hashtable globalTable = new Hashtable();
    
    
    /** 
     * Special prototype instance for native side
     * calling pseudo instance methods.
     */
    /*friendly*/ native void reportIluOIntInst();
    /*friendly*/ static IluOInt iluOIntProto = null;
    public static void init() {
        if (iluOIntProto != null) return;
        if (! IluInit.doInit()) return;
        iluOIntProto = new IluOInt();
        iluOIntProto.reportIluOIntInst();
    } //init
    
    static {
        init();
    } //static

    /** No random creations */
    private IluOInt() {
        super();
        if (IluDebug.dontGC() != 0) {
            //Prevent any garbage collection.
            //Temporary courtesy to early users until
            //finalization is trusted to work correctly.
            IluForEver.remember(this);
        }
        if (IluDebug.traceObjects() > 0) {
            IluDebug.log.println("! IluOInt: new " + this);
        }
    }
    
    /** 
     * Reserved for LSR to generate new instances.
     * Instance is not yet created all the way.
     */
    /*friendly*/ final IluOInt allocateOI() {
        //RT will fill in fields !
        return new IluOInt();
    } //allocateOI
    
    /**
     * Returns the object of this IluOInt.
     * Finishes allocation if necessary     
     */
    public final xerox.ilu.IluObject getTheObject() {
        xerox.ilu.IluObject obj = this.jjJavaObject;
        if (obj==null) {
            //if the object exists it can be true or surrogate
            //but we only create surrogates here
            this.createSurrogateForOI();
            obj = this.jjJavaObject;
        }
        return obj;
    } //getTheObject
    
    //tables to temporarily prevent gc
    private static Hashtable interestTable = new Hashtable(); 
    
    //Finalize real work on harmless thread
    private static IluOIntFinalizer finThread = new IluOIntFinalizer();

    /** 
     * To prevent garbage collection of true object 
     */
    public static
    void retain(xerox.ilu.IluObject realObject) {
        IluOInt oi = peek(realObject);
        if (oi == null ) throw new IluRuntimeError("can't retain");
        oi.retainOI();
    } //retain
    
    /*
     * To prevent garbage collection of true object.
     * Object will not be garbage collected as long as server is alive.  
     */
    /*friendly*/ void retainOI() {
        if (this.destroyed) {
            throw new IluRuntimeError("destroyed; too late to retain");
        }
        xerox.ilu.IluServer server = this.jjServer;
        if (server != null) {
            if (server.surrogateServer) {return;}
            if (IluDebug.traceGC()>2) {
                IluDebug.log.println("! IluOInt.retain " + this);
            }
            this.retained = true;
            server.retainTable.put(this, this);
        }
    } //retainOI
    
    /** 
     * Allow garbage collection of true object 
     */
    public static
    void expose(xerox.ilu.IluObject realObject) {
        IluOInt oi = peek(realObject);
        if (oi != null) oi.exposeOI();
    } //expose
    
    /*
     * Allow garbage collection of true object 
     */
    /*friendly*/ void exposeOI() {
        xerox.ilu.IluServer server = this.jjServer;
        if (server != null) {
            if (server.surrogateServer) {return;}
            if (IluDebug.traceGC()>2) {
                IluDebug.log.println("! IluOInt.expose " + this);
            }
            this.retained = false;
            server.retainTable.remove(this);
        }
    } //exposeOI
    
    /* 
     * Called from from kernel when it starts being "very interested". 
     * Server lock is/must-be held
     */
    /*friendly*/ void showInterest() {
        if (IluDebug.traceGC() > 0) {
            IluDebug.log.println("! IluOInt.showInterest " + this);
        }
        this.veryInterest = true;
        interestTable.put(this, this);
        if (! this.retained) this.ghost = true;
    } //showInterest
    
    /* 
     * Called from from kernel when it is no more "very interested". 
     * Server lock is/must-be held
     */
    /*friendly*/ void removeInterest() {
        if (IluDebug.traceGC() > 0) {
            IluDebug.log.println("! IluOInt.removeInterest " + this);
        }
        this.veryInterest = false;
        interestTable.remove(this);
    } //removeInterest
    
    
    /** 
     * Given a prospective real TRUE object, create the internal 
     * object. A real object may have at most one internal.
     *
     * Throws exception if object already known to ilu.
     */
    public static 
    IluOInt newOI(
                xerox.ilu.IluObject realObject, 
                xerox.ilu.IluClassRep iluClassRep,
                xerox.ilu.IluServer server
                ) throws xerox.ilu.IluSystemException {
        if (realObject==null) {
            throw new IluRuntimeError("null object");
        }
        IluOInt oi = iluOIntProto.allocateOI();
        oi.jjClassRep = iluClassRep;
        oi.jjServer = server;
        oi.jjJavaObject = realObject;
        if (realObject instanceof IluDataSupport) {
            IluDataSupport ids = (IluDataSupport) realObject;
            synchronized (realObject) {
                if (ids.getIluData() != null) {
                    IluDebug.clientPreError("object already exists");
                    throw new IluRuntimeError("object already exists");
                }
                ids.setIluData(oi);
            }
        } else {
            synchronized (globalTable) {
                if (globalTable.get(realObject) != null) {
                    IluDebug.clientPreError("object already exists");
                    throw new IluRuntimeError("object already exists");
                }
                globalTable.put(realObject, oi);
            }
        }
        if (IluDebug.traceObjects() > 0) {
            IluDebug.log.println("! IluOInt oi: " + oi + " obj: " + realObject);
        }
        return oi;
    } //newOI
    
    /* undo the effect of newOI in case native registration failed ... */
    /*friendly*/ void forgetOI() {
        xerox.ilu.IluObject realObject = this.jjJavaObject;
        if (realObject !=null) {
            globalTable.remove(realObject);
        }
    }
        
    public int proposeLifeTime(int proposal) {
        if (proposal == IluLifetimeArgs.iluLifetimeRemember) return proposal;
        if (proposal == IluLifetimeArgs.iluLifetimeForget) return proposal;
        if (this.jjJavaObject != null) {
            if (this.jjJavaObject instanceof IluDataSupport) {
                return IluLifetimeArgs.iluLifetimeForget;
            }
        }
        if (this.jjClassRep != null) {
            if (this.jjClassRep.isCollectible()) {
                return IluLifetimeArgs.iluLifetimeForget;
            }
        }
        return IluLifetimeArgs.iluLifetimeRemember;
    } //proposeLifeTime
    
    /** 
     * Given a prospective real TRUE object, find an internal if it exists.
     * Returns null if none exists or existence can not be proved.
     *
     * (If caller is outside the server lock, the Ilu'ness of the returned
     * IluOInt can get lost, but the memory is safe).
     */
    public static IluOInt
    peek(xerox.ilu.IluObject realObject) {
        if (IluDebug.callsFlag > IluDebug.basic) {
            IluDebug.log.println("! Enter IluOInt.peek, from " + realObject);
        }
        if (realObject instanceof IluDataSupport) {
            IluDataSupport ids = (IluDataSupport) realObject;
            java.lang.Object raw = ids.getIluData();
            if (IluDebug.callsFlag > IluDebug.detailed) {
                IluDebug.log.println("! Exit IluOInt.peek, supp > " + raw );
            }
            return (IluOInt) raw; 
        }
        IluOInt oi = (IluOInt) globalTable.get(realObject);
        if (IluDebug.callsFlag > IluDebug.basic) {
            IluDebug.log.println("! Exit IluOInt.peek, table > " + oi );
        }
        return oi;
    } //peek

    /** 
     * Called by LSR to finish generating new SURROGATE instances.
     * This is usefull for SURROGATE objects.
     * It can be called for true objects, but since jjJavaObject is 
     * not null there will be no object creation.
     */
    private void createSurrogateForOI() 
               throws xerox.ilu.IluSystemException {
        if (IluDebug.traceObjects() > 1) {
            IluDebug.log.println("! IluOInt: enter createSurrogateForOI oi: " 
                + this + " obj: " + this.jjJavaObject);
        }
        synchronized (this) {
            if (this.jjJavaObject != null) {
                //corresponding real object has already been created 
                return;
            }
            xerox.ilu.IluObject surr = null;
            if (this.jjClassRep.factory != null) {
               java.lang.Object stupidCast = this;
               surr = (xerox.ilu.IluObject)
                   this.jjClassRep.factory.createSurrogateObject(stupidCast);
            } else if (this.jjClassRep.surrogateJClass != null) {
                try {
                    surr = (xerox.ilu.IluObject)
                        this.jjClassRep.surrogateJClass.newInstance();
                } catch (IllegalAccessException e) {
                } catch (InstantiationException e) {
                }
            }
            if (surr == null) {
                IluDebug.clientPreError("failed creating surrogate");
                throw new IluRuntimeError("failed creating surrogate");
            }
            this.jjJavaObject = surr;
            if (surr instanceof IluDataSupport) {
                 ((IluDataSupport)surr).setIluData(this);
            } else {
                synchronized (globalTable) {
                    if (globalTable.get(surr) != null) {
                        IluDebug.clientPreError("object already exists");
                        throw new IluRuntimeError("object already exists");
                    }
                    globalTable.put(surr, this);
                }
            }
        }
        if (IluDebug.traceObjects() > 1) {
            IluDebug.log.println("! IluOInt:  done createSurrogateForOI oi: " 
                + this + " obj: " + this.jjJavaObject);
        }
    } //createSurrogateForOI
    
    /** 
     * Start eventually forgetting object...
     */
    protected final void  
    finalize() throws java.lang.Throwable {
        if (IluDebug.traceObjects() > 1) {
            IluDebug.log.println("! IluOInt: finalize " + this
                + " destr." + this.destroyed);
        }
        if (this.destroyed) {
            //we are not interested in this in anymore
            return;
        }
        //use separate thread so we won't lock up the real finalizer
        finThread.provide(this);
    } //finalize
    
    /** 
     * Report finalization to IluWPBase class...
     */
    /*friendly*/ final void
    superFinalize() throws java.lang.Throwable {
        super.finalize();
    }
    
    
    /** 
     * "Destroys object" if it is neither used or appears to be used.
     * Might or might not succeed.
     * Returns whether there is a need to re-finalize
     */ 
    /*friendly*/ native boolean destroyGentleOI();
    
    /** 
     * "Destroys object" and ignores whether the application keeps using it.
     * Further use of the object might raise whatever errors in the 
     * application; but the kernel will keep its piece of the object 
     * until it is safe. You get what you deserve.
     */ 
    /*friendly*/ void destroyRudeOI() {
         IluObject ob0 = this.jjJavaObject;
         if (ob0 != null) {
             if (!(ob0 instanceof IluDataSupport)) {
                 synchronized(globalTable) {globalTable.remove(ob0);}
             }
         }
         this.exposeOI();
         this.nDestroyRudeOI();
    } //destroyRudeOI
    private native void nDestroyRudeOI();

    /** 
     * Publishes the SBH of the object in the local object domain
     */
    /*friendly*/ native void publishOI();
    
    /** 
     * withdrawOI is undo of publishOI
     */
    /*friendly*/ native void withdrawOI();
    
    /** 
     * See comments in IluRT0.sbhOfObject
     */
    /*friendly*/ native java.lang.String sbhOfOI();
    
    /** 
     * Returns OMG IIOP-specified IOR string for object.
     * May return null if object is not exported through IIOP.
     */
    /*friendly*/ native java.lang.String iorOfOI();
    
    /** 
     * See comments in IluRT0.ping
     */
    /*friendly*/ native void nativePingOI();
    
    /** 
     * See comments in IluRT0.URLOfObject
     */
    /*friendly*/ native java.lang.String nativeURLOfObject();
    
    public static int longNames = 1;
    
    /** 
     * Standard procedure overridden to include ilu kernel obj.
     */
    public java.lang.String toString() {
        xerox.ilu.IluObject ob = this.jjJavaObject;
        java.lang.String s =
            "ilu-oi@" + java.lang.Integer.toHexString(this.hashCode()) + "<" 
            + java.lang.Long.toHexString(this.yIluKernelObject) + ">";
        if ((ob!=null) && (longNames>0)) {
            //don't use ob.toString() since some object 
            //might recurse and print IluOInt...
            s = s + "(" + java.lang.Integer.toHexString(ob.hashCode()) +")";
        }
        return s;
    } //toString
    

} //IluOInt


/*friendly*/ final class IluOIntFinalizer 
    extends xerox.basics.Consumer0 {

     private static int count = 0; // used to provide unique thread name
     private static IluFinalizeReporter reporter = 
         new IluFinalizeReporter();

    /* returns and increments the classes object count */
    private static final synchronized long incrementCount() {
        return count++;
    }

    /*friendly*/ IluOIntFinalizer() {
        super(IluEnvironment.threadPriority(), "IluOIntFinalizer" + incrementCount());
    } //IluOIntFinalizer
    
    protected void consume(java.lang.Object obj) {
        //take object from finalization queue 
        //and decide whether to finalize it or not (yet) 
        IluOInt oi = (IluOInt) obj;
        if (IluDebug.traceObjects() > 0) {
            IluDebug.log.println("! IluOIntFinalizer: consume " 
                + oi + " (" + oi.jjJavaObject + ")"
                );
        }
        if (oi.destroyed) {
            //do nothing; the ilu object has already been destroyed
            if (IluDebug.traceObjects() > 0) {
                IluDebug.log.println("! IluOIntFinalizer: already dead"  + oi);
            }
        } else if (oi.ghost) {
            //de-ghost object
            if (IluDebug.traceObjects() > 0) {
                IluDebug.log.println("! IluOIntFinalizer: de-ghost " + oi);
            }
            oi.ghost = false;
            IluReEnableFinalization.reEnable(oi);
        } else {
            //try to destroy
            boolean need_to_re_finalize = oi.destroyGentleOI();
            if (need_to_re_finalize) {
                //destroy failed; try again later
                if (IluDebug.traceObjects() > 0) {
                    IluDebug.log.println("! IluOIntFinalizer: try later " + oi);
                }
                oi.ghost = false;
                IluReEnableFinalization.reEnable(oi);
            } else {
                //destroy succeeded; optionally report it to client
                if (IluDebug.traceObjects() > 0) {
                    IluDebug.log.println("! IluOIntFinalizer: done " + oi);
                }
                xerox.ilu.IluObject robj = oi.jjJavaObject;
                if (robj instanceof xerox.ilu.IluServantFinalizable) {
                    reporter.provide(robj);
                }
                //handle deferred IluWPBase finalization
                try {
                    oi.superFinalize();
                } catch (java.lang.Throwable t) {
                }
            }
        } 
    } //consume

} //IluOIntFinalizer


/*friendly*/ final class IluFinalizeReporter 
    extends xerox.basics.Consumer0 {

    private static int count = 0; // used to provide unique thread name

    /* returns and increments the classes object count */
    private static final synchronized long incrementCount() {
        return count++;
    }

    /*friendly*/ IluFinalizeReporter() {
        super(IluEnvironment.threadPriority(), "IluFinalizeReporter" + incrementCount());
    } //IluFinalizeReporter
    
    protected void consume(java.lang.Object obj) {
        if (IluDebug.traceObjects() > 1) {
            IluDebug.log.println("! IluFinalizeReporter: consume " + obj);
        }
        xerox.ilu.IluServantFinalizable fobj =
            (xerox.ilu.IluServantFinalizable) obj;
        try {
            fobj.iluServantFinalize(); 
        } catch (java.lang.Throwable t) {
        }
    } //consume

} //IluFinalizeReporter


/*friendly*/ final 
class IluReEnableFinalization {
    private IluOInt theOI = null;
    /*friends*/ static void reEnable(IluOInt oi) {
        new IluReEnableFinalization(oi);
    }
    /*friends*/ IluReEnableFinalization(IluOInt oi) {
        theOI = oi;
        IluReEnableFinalization old = oi.reFinalizer;
        if (old != null) old.theOI = null;
        oi.reFinalizer = this;
        if (IluDebug.traceObjects() > 0) {
            IluDebug.log.println("! IluReEnableFinalization reenable " + oi);
        }
    }
    protected void finalize() throws Throwable {
        IluOInt oi = theOI;
        if (IluDebug.traceObjects() > 0) {
            IluDebug.log.println("! IluReEnableFinalization finalize " + oi);
        }
        if (oi != null) oi.finalize();
    } 
} //IluReEnableFinalization


