/* IluSystemException0.java */
/* Chris Jacobi, April 21, 1997 10:31 am PDT */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */ 
 

/*
 * $Id: IluSystemException0.java,v 1.6 1997/04/21 18:45:41 jacobi Exp $
 */

package xerox.ilu;

/**
 * IluSystemException0 is an artificial superclass of IluSystemException
 * - Clients may use this to catch all ilu system exceptions,
 *   but since all system exceptions are also IluSystemException's,
 *   clients can as well totally ignore this class.
 * - Clients must not throw this exceptions. 
 * - Clients must not declare methods to throw this exceptions. 
 */
public class IluSystemException0 extends xerox.ilu.IluSystemExceptionBase {

    public IluSystemException0() {
        super();
    }

    public IluSystemException0(java.lang.String s) {
        super(s);
    }
    
    static public final int completedYes = 0;
    static public final int completedNo = 1;
    static public final int completedMaybe = 2;
    
    private int theCompleted = completedMaybe;
    private int theMinor = 0;
    
    public int getCompleted() { 
        return theCompleted;
    }
    
    public xerox.ilu.IluSystemException setCompleted(int completed) {
        if (completed>completedMaybe) completed = completedMaybe;
        theCompleted = completed;
        return (xerox.ilu.IluSystemException) this;
    }

    public int getMinor() { 
        return theMinor;
    }
    
    public xerox.ilu.IluSystemException setMinor(int minor) { 
        theMinor = minor;
        return (xerox.ilu.IluSystemException) this;
    }
    
    public java.lang.String toString() {
        java.lang.String s = super.toString();
        switch (theCompleted) {
            case completedMaybe:
                break;
            case completedYes:
                s = s + " completed: yes";
                break;
            case completedNo:
                s = s + " completed: no";
                break;
            default:
                s = s + " completed: bad";
                break;
        }
        if (theMinor  != 0) {
            s = s + " minor code: " + theMinor;
        }
        return s;
    }


} // IluSystemException0
