/* hlpcls.c */
/* Chris Jacobi, August 18, 1997 8:05 pm PDT */

/*
 * Copyright (c) 1991-1996 Xerox Corporation.  All Rights Reserved.
 * Unlimited use, reproduction, and distribution of this software is permitted.
 * Any copy of this software must include both the above copyright notice of
 * Xerox Corporation and this paragraph.  Any distribution of this software
 * must comply with all applicable United States export control laws.  This
 * software is made available AS IS, and XEROX CORPORATION DISCLAIMS ALL
 * WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
 * NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY LIABILITY FOR
 * DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED,
 * WHETHER ARISING IN CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT
 * LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/* $Id: hlpcls.c,v 1.15 1997/08/20 21:52:43 jacobi Exp $ */

#include <stdio.h>
#include <string.h>
#include "iluptype.h"

#include "shared.h"
#include "stubops.h"
#include "io.h"
#include "name.h"
#include "util.h"
#include "context.h"
#include "genobj.h"
#include "genenum.h"
#include "genopt.h"



#define TYPECODE "xerox.ilu.IluTypeCode"
#define THROWSYS "throws xerox.ilu.IluSystemException"


PRIVATE void 
print_IOF_ImplComment()
{
    printf("    /** Not public; implements xerox.ilu.IluIOFunctions */\n");
}


PUBLIC char*
helperClassShortName(Type rawT)
{
    IHandle ih = getContextT(rawT);
    char* classBaseName = easyShortTypeNameCleaned(rawT); 
    char* classShortName = cat2(classBaseName, ih->p.helperSuffix);
    return classShortName;
}



PRIVATE void
printHelperIOMethods(Type urT)
{
    /* for objects suboptimal: Could share impl and use tc to find class */
    print_IOF_ImplComment();
    printf("    public int szFunc(xerox.ilu.IluCall _call, java.lang.Object _x, " TYPECODE " _tc) " THROWSYS " {\n");
    printf("        %s _xx = %s;\n", 
        typeDeclarator(urT), 
        fromObject(urT, "_x")
        );
    printf("        return %s;\n", ioSzPiece(urT, "_xx"));
    printf("    } //szFunc \n\n");
    
    print_IOF_ImplComment();
    printf("    public void outFunc(xerox.ilu.IluCall _call, java.lang.Object _x, " TYPECODE " _tc) " THROWSYS " {\n");
    printf("        %s _xx = %s;\n", 
        typeDeclarator(urT), 
        fromObject(urT, "_x")
        );
    printf("        %s;\n", ioOutPiece(urT, "_xx"));
    printf("    } //outFunc \n\n");

    print_IOF_ImplComment();
    printf("    public java.lang.Object inFunc(xerox.ilu.IluCall _call, " TYPECODE " _tc) " THROWSYS " {\n");
    printf("        return %s;\n", toObject(urT, ioInPiece(urT)));
    printf("    } //inFunc \n\n");
    
    printf("    public boolean isAFunc(java.lang.Object _x, " TYPECODE " _tc) {\n");
    printf("        if (_x == null) return %s;\n", booleanImage(opt_is_a(urT)));
    printf("        return (_x instanceof %s);\n", wrapperTypeDeclarator(urT));
    printf("    } //isAFunc \n\n");
}


PUBLIC void
printHelperClass(Type rawT)
{
    /* ONE OF THE FEW PLACES DEALING WITH ALIAS */
    char* classShortName;
    Type urT;
    IHandle rawIH = getContextT(rawT);
    
    if (! rawIH->p.genHlp) return;
    
    if (type_kind(rawT) == alias_Type) {
        urT = myUrType(rawT);
    } else {
        urT = myUrType(rawT);
        rawT = urT;
    }
    
    classShortName = helperClassShortName(rawT);
    NewJavaFile(
        rawT->interface, 
        getContext(rawT->interface, rawT->scoping), 
        classShortName);
    LoadJavaClass(classShortName);
    if (obj_is_a(rawT)) {
        Class c = class_object(rawT);
        if (c->doc_string) printDocString("", c->doc_string);
    
    }
    printf("public class %s %s {\n",
        classShortName,
        "implements xerox.ilu.IluIOFunctions "
        );
    
    printf("    private final static java.lang.String _id = \n        %s;\n",
        qoString(rawT->uid)
        );
    printf("    private static " TYPECODE " _tc = null;\n");
    printf("    static {\n");
    printf("        _tc = " TYPECODE ".newTypeCode(id(), %s, %s);\n",
        cat3("new ", classShortName, "()"),
        corbaTCKind(type_kind(rawT))
        );
    printf("    }\n\n");
    
    printf("    public final static java.lang.String id() {\n");
    printf("        return _id;\n");
    printf("    } //id \n\n");
    
    printf("    public final static " TYPECODE " type() {\n");
    printf("        return _tc;\n");
    printf("    } //type \n\n");
    
    if (obj_is_a(rawT)) {
        printf("    public static %s narrow(xerox.ilu.IluObject _x) {\n",
            typeDeclarator(urT)
            );
        printf("        return (%s) _x;\n", typeDeclarator(urT));
        printf("    } //narrow \n\n");
    }
    
    printf("    public static void insert(org.omg.CORBA.Any _any, %s _x) " THROWSYS " {\n",
        typeDeclarator(urT)
        );
    printf("        xerox.ilu.IluPickle _pickle = to_pickle(_x);\n");
    printf("        ((xerox.ilu.IluAny)_any).setPickle(_pickle);\n");
    printf("    } //insert \n\n");

    printf("    public static %s extract(org.omg.CORBA.Any _any) " THROWSYS " {\n",
        typeDeclarator(urT)
        );
    printf("        xerox.ilu.IluPickle _pickle = ((xerox.ilu.IluAny)_any).getPickle();\n");
    printf("        return from_pickle(_pickle);\n");
    printf("    } //extract \n\n");

    printf("    public static xerox.ilu.IluPickle to_pickle(%s _x) " THROWSYS " {\n",
        typeDeclarator(urT)
        );
    printf("        int _sz = 0;\n");
    printf("        xerox.ilu.IluCall _call = null;\n");
    printf("        xerox.ilu.IluPickle _pickle = new xerox.ilu.IluPickle();\n");
    printf("        try {\n");
    printf("            _call = _pickle.startToPickle();\n");
    printf("            _sz = %s;\n", ioSzPiece(urT, "_x"));
    printf("            _pickle.midToPickle(_call, _sz, id());\n");
    printf("            %s;\n", ioOutPiece(urT, "_x"));
    printf("        } finally {\n");
    printf("            _pickle.endToPickle(_call);\n");
    printf("        }\n");
    printf("        return _pickle;\n");
    printf("    } //to_pickle \n\n");

    printf("    public static %s from_pickle(xerox.ilu.IluPickle _pickle) " THROWSYS " {\n",
        typeDeclarator(urT)
        );
    printf("        %s _x = %s;\n", 
        typeDeclarator(urT), 
        typeInitializer(type_kind(urT))
        );
    printf("        xerox.ilu.IluCall _call = null;\n");
    printf("        try {\n");
    printf("            _call = _pickle.startFromPickle(id());\n");
    printf("            _x = %s;\n", ioInPiece(urT));
    printf("        } finally {\n");
    printf("            _pickle.endFromPickle(_call);\n");
    printf("        }\n");
    printf("        return _x;\n");
    printf("    } //from_pickle \n\n");
    
    if (1) {
        printHelperIOMethods(urT);
    }

    printf("}//%s\n\n", classShortName);
}


/* end */
